from xml.etree.ElementTree import tostring

from yarok.core.components_manager import ComponentsManager
from yarok.core.config_block import ConfigBlock
from yarok.core.platform import Platform


class PlatformHW(Platform):
    """
        Is the Platform for the Real World

    """

    def __init__(self, manager: ComponentsManager, config: ConfigBlock, behaviour):
        super(PlatformHW, self).__init__(manager, config, behaviour)
        self.interfaces = {
            id: self.manager.config(id)['interface_hw'](
                **{'config': config['interfaces'][c['name_path']] or ConfigBlock({})})
            for id, c in self.manager.components.items()
            if 'interface_hw' in self.manager.config(id)}

        self.init_components(self.interfaces)

        while not self.interfaces_ready():
            self.step()

    def interfaces_ready(self):
        return not (sum(
            [not self.interfaces[i].is_ready() for i in self.interfaces if hasattr(self.interfaces[i], 'step')]) > 0)

    def step(self):
        """
            Steps through all the interfaces
        """
        _ = {self.interfaces[i].step() for i in self.interfaces if hasattr(self.interfaces[i], 'step')}



from yarok.core.components_manager import component
from yarok.comm.components.cam.cam import Cam
from yarok.platforms.mjc.interface import InterfaceMJC

import cv2
import os
import numpy as np

from .model.simulation_model import SimulationApproach

class GelSight2014InterfaceMJC:

    def __init__(self, interface: InterfaceMJC):
        self.interface = interface
        self.raw_depth = np.zeros((480, 640))
        self.tactile_rgb = np.zeros((480, 640))

        self.approach = SimulationApproach(**{
            'light_sources': [
                {'position': [0, 1, 0.25], 'color': (255, 255, 255), 'kd': 0.6, 'ks': 0.5},  # white, top
                {'position': [-1, 0, 0.25], 'color': (255, 130, 115), 'kd': 0.5, 'ks': 0.3},  # blue, right
                {'position': [0, -1, 0.25], 'color': (108, 82, 255), 'kd': 0.6, 'ks': 0.4},  # red, bottom
                {'position': [1, 0, 0.25], 'color': (120, 255, 153), 'kd': 0.1, 'ks': 0.1}  # green, left

                # {'position': [-1, 0, 0.25], 'color': (255, 255, 255), 'kd': 0.5, 'ks': 0.3},  # white, top
                # {'position': [0, 0, 1], 'color': (255, 255, 255), 'kd': 0.5, 'ks': 0.3},  # blue, right

            ],
            'background_img': cv2.imread(os.path.dirname(__file__) + '/background_gelsight2014.png'),
            'ka': 0.8,
            'px2m_ratio': 5.4347826087e-05,
            'elastomer_thickness': 0.004,
            'min_depth': 0.03,
            'texture_sigma': 0.000002
        })

    def read(self, shape=(480, 640)):
        self.raw_depth = self.interface.read_camera('camera', shape, True)[1]
        self.tactile_rgb = self.approach.generate(self.raw_depth)
        return self.tactile_rgb

    def read_depth(self):
        return self.raw_depth

@component(
    components=[
        Cam
    ],
    interface_mjc=GelSight2014InterfaceMJC,
    probe=lambda c: {'gelsight camera': c.read()},
    # language=xml
    template="""
<mujoco>
    <asset>
        <material name="black_resin" rgba="0.1 0.1 0.1 1"/>
        <material name="gray_elastomer" rgba="0.8 0.8 0.8 1"/>
        <material name="transparent_glass" rgba="0.9 0.95 1 0.7"/>

        <!-- gelsight models-->
        <mesh name="gelsight_front" file="meshes/gelsight2014_front.stl" scale="0.001 0.001 0.001"/>
        <mesh name="gelsight_back" file="meshes/gelsight2014_back.stl" scale="0.001 0.001 0.001"/>
        <mesh name="gelsight_glass" file="meshes/glass.stl" scale="0.0009 0.00125 0.001"/>

        <mesh name="front_cover" file="meshes/mountable_gelsight.stl" scale="0.001 0.001 0.001"/>
        <mesh name="back_cover" file="meshes/back_cover.stl" scale="0.001 0.001 0.001"/>

    </asset>
    <worldbody>
        <body name="sensor_body" pos="0 0 0">
            <!--Front and Back-->
            <geom name="front" type="mesh" material="black_resin" mesh="front_cover" mass="0.05" contype="32"
                  conaffinity="32" friction="1 0.05 0.01" solimp="1.1 1.2 0.001 0.5 2" solref="0.02 1"/>
            <geom name="back" type="mesh" material="black_resin" mesh="back_cover" mass="0.05" contype="32"
                  conaffinity="32" friction="1 0.05 0.01" solimp="1.1 1.2 0.001 0.5 2" solref="0.02 1"/>

            <!-- Glass Cover-->
            <geom name="glass0" type="mesh" material="transparent_glass" mesh="gelsight_glass" mass="0.005"
                  contype="32" conaffinity="32" pos="-0.011 0 0.029"/>
            <geom name="glass1" type="mesh" material="transparent_glass" mesh="gelsight_glass" mass="0.005"
                  contype="32" conaffinity="32" pos="0.0115 0 0.029" quat="0 0 0 1"/>

            <!-- Elastomer -->
            <geom name="elastomer" type="box" size="0.013 0.013 0.001" euler="0 0 0" pos="0 0 0.033"
                  contype="0"
                  conaffinity="32" rgba="0.9 0.95 1 0.1"/>

            <!-- Elastomer Cover -->
            
            <geom name="elast_cover" type="box" size="0.013 0.013 0.00001" euler="0 0 0" pos="0 0 0.034001"
                  contype="0" conaffinity="32" material="black_resin"
                  friction="1 0.05 0.01" solimp="1.1 1.2 0.001 0.5 2" solref="0.02 1"/>

            <!-- Gel Camera -->
            <camera name="camera" mode="fixed" pos="0 0 0.001" zaxis="0 0 -1" fovy="20"/>

            <!-- Friction placholder -->
            <geom name="friction" type="box"
                 size="0.013 0.013 0.00001"
                 euler="0 0 0"
                 pos="0 0 0.032"
                 contype="32"
                 conaffinity="32" rgba="0 0 0 0"
                 friction="1 0.05 0.01" solimp="1.1 1.2 0.001 0.5 2" solref="0.02 1"/>
        </body>
    </worldbody>
</mujoco>
    """
)
class GelSight2014:

    def __init__(self):
        """
            GelSight 2014 driver as proposed in "Generation of GelSight Tactile Images for Sim2Real Learning＂
            https://danfergo.github.io/gelsight-simulation/

            The frame method gets the depth map from the simula
        """

    def read(self):
        pass


import copy
import pathlib
import re
import xml.etree.ElementTree as ET
from xml.etree.ElementTree import tostring

from yarok.core.config_block import ConfigBlock

# attributes that are renamed from 'name' to 'id'
RENAME_KEYS_ATTRS_NAMES = [
    'name',
    'joint',
    'actuator',
    'body1',
    'body2',
    'joint1',
    'joint2',
    'material',
    'mesh',
    'texture',
    'tendon'
]

# names of the blocks to be merged
# except worldbody that follows a special merging process
# i.e. nesting components, etc.
EXTENDABLE_BLOCKS_NAMES = [
    'default',
    'asset',
    'tendon',
    'sensor',
    'actuator',
    'equality',
    'contact'
]


def stringify(el):
    return tostring(el, encoding="unicode")


def stringify_header(el):
    """ used to construct exception messages """
    """ returns  """
    return tostring(el, encoding="unicode") \
               .replace('\n', '') \
               .strip() \
               .split('>')[0] + '>'


class ComponentsManager:
    """
    Handles loading components templates (.xml) & controllers (.py)
    and Builds xml tree and components i.e.,

    - self.xml_tree   =>
    - self.components => dict ([component.id] => dict(
        class => component factory class,
        name => component name,
        tag => component tag name (to be used in the template)
        class_name => component factory (snake-case) name,
        id => auto generated id,
        children => [dict(...)],
        instance => object ref             # set while instantiating the components
    ))


    """

    def __init__(self, root_component, components_config: ConfigBlock):
        self.id_counter = 0
        # used for direct access to component
        self.components_config = components_config
        self.components = {}

        # used to iterate through all components, depth first,
        # leafs-to-root, because of dependencies, etc.
        config = components_config['/'] if '/' in components_config else ConfigBlock({})
        config.defaults(root_component.__data__['defaults'])
        self.components_tree = {
            'class': root_component,
            'name': '',
            'name_path': '/',
            'id': self.gen_id(),
            'parent': None,
            'children': [],
            'config': config
        }
        self.xml_tree = self.load_tree(self.components_tree)

        if True:
            f = open("scene.xml", "w")
            f.write(tostring(self.xml_tree, encoding='unicode'))
            f.close()

        # build components map, for direct access
        def walk_components(component):
            for child in component['children']:
                walk_components(child)
            self.components[component['id']] = component

        walk_components(self.components_tree)

    def is_id(self, idx):
        return isinstance(idx, int) or isinstance(idx, str)

    def gen_id(self, name=''):
        self.id_counter += 1
        # '__c__' + str(self.id_counter) + '-' +
        return name + '#' + str(self.id_counter)

    def get(self, comp_id):
        return self.components[comp_id]

    def get_by_name(self, c, name):
        """
            Gets a component by 'name', given its parent 'c' (or its parent id)
        """
        comp = self.components[c] if self.is_id(c) else c
        for child in comp['children']:
            if name == child['name']:
                return child
        return None

    def data(self, c):
        comp = self.components[c] if self.is_id(c) else c
        return comp['class'].__data__

    def config(self, c):
        comp = self.components[c] if self.is_id(c) else c
        return comp['config']

    def load_tree(self, comp):
        """
            The method that loads the multiple components and its template (.xml) into
            a single xml/model tree. It also loads the multiple components into a tree
        """

        # inits auxiliar variables from the component metadata
        template = comp['class'].__data__['template']
        xml_path = comp['class'].__data__['template_path']
        id = comp['id']

        # loads xml tree from xml path
        tree = ET.fromstring(template) if template is not None else ET.parse(xml_path).getroot()

        # handling lean trees
        # if tree.tag != 'mujoco':
        #     mujocoRoot = ET.Element("mujoco")
        #     mujocoRoot.append(tree)
        #     tree = mujocoRoot

        # if not any([e.tag =='worldbody' for e in tree.getChildren()]):
        #     worldbodyRoot = ET.Element("worldbody")
        #     worldbodyRoot.append(tree)

        extends = comp['class'].__data__['extends']
        if extends is not None:
            cls = comp['class']
            comp['class'] = extends
            ext_tree = self.load_tree(comp)
            # element = ET.Element("ext-wrapper")
            # wb = ext_tree.find('worldbody')
            self.merge_nonbody_blocks(ext_tree, tree)
            ext_tree.find('worldbody').extend(tree.find('worldbody'))

            comp['class'] = cls
            tree = ext_tree

        # re-pathing all file refs, that are not absolute refs,
        # to be relative to the .xml file
        dir_path = str(pathlib.Path(xml_path).parent.resolve())
        for e in tree.iter():
            if 'file' in e.attrib and not e.attrib['file'].startswith('/'):
                e.attrib['file'] = dir_path + '/' + e.attrib['file']

        # renaming primary&foreign keys to from xxx to id:xxx
        # !!!! understand the rename_fk in two places. is it simplifiable?
        def rename_fk_recur(els):
            for el in els:
                if el.tag == 'worldbody':
                    continue
                try:
                    self.parse_attributes(el, comp['config'])
                except Exception as e:
                    c_path = comp['name_path']
                    raise Exception('\n\n' + e.args[0]
                                    + ',\nfor ' + c_path
                                    + ',\n\nplease patch yarok config with { components: { ' + c_path + ' : { ' +
                                    e.args[
                                        1] + ': <value> }}}  '
                                    )

                for attr in RENAME_KEYS_ATTRS_NAMES:
                    if attr in el.attrib:
                        el.attrib[attr] = id + ':' + el.attrib[attr]
                rename_fk_recur(list(el))

        rename_fk_recur(list(tree))

        # walking current file .xml tree and looking for components and instantiating them
        def walk_tree_recur(parent, elements, parent_component, config=None):
            c_data = parent_component['class'].__data__

            for element in elements:
                tag = element.tag

                try:
                    self.parse_attributes(element, config or comp['config'])
                except Exception as e:
                    raise Exception(e.args[0] + ',\nfor ' + str(comp))

                has_if = 'if' in element.attrib
                if has_if:
                    condition_str = element.attrib['if']
                    element.attrib.pop('if')
                    condition = eval(condition_str, None, config or parent_component['config'])
                    if not condition:
                        parent.remove(element)
                        continue

                if tag in EXTENDABLE_BLOCKS_NAMES:
                    continue

                elif tag == 'for':
                    condition_str = element.attrib['each']
                    var_name = element.attrib['as'] if 'as' in element.attrib else None
                    # print('--> ', parent_component['config'],  globals()['range'])
                    conf = config or parent_component['config']
                    elems = eval(condition_str, None, conf)
                    parent.remove(element)

                    for e in elems:
                        new_config = None if var_name is None else ConfigBlock({
                            var_name: e
                        })
                        new_config.defaults(config or comp['config'])
                        new_element = copy.deepcopy(element)
                        parent.extend(new_element)
                        walk_tree_recur(parent, list(new_element), comp, config=new_config)

                elif tag == 'if':
                    condition_str = element.attrib['the']
                    condition = eval(condition_str)
                    parent.remove(element)

                    if condition:
                        parent.extend(element)
                        walk_tree_recur(parent, list(element), comp)

                elif tag in c_data['components']:
                    component_tag = tag
                    component_name = element.attrib['name']

                    if component_name == '':
                        raise KeyError(
                            'No component should be named ""!, ' +
                            component_tag +
                            '[name=root] at' +
                            parent_component['name'])

                    # creates new sub component data wrapper
                    sub_component = {
                        'id': self.gen_id(component_name),
                        'name': component_name,
                        'name_path': ('/' + component_name) \
                            if (parent_component['name_path'] == '/') \
                            else (parent_component['name_path'] + '/' + component_name),
                        'tag': component_tag,
                        'class': c_data['components'][component_tag],
                        'parent': parent_component,
                        'children': []
                    }
                    config = self.components_config[sub_component['name_path']] \
                        if sub_component['name_path'] in self.components_config \
                        else ConfigBlock({})
                    config.defaults(sub_component['class'].__data__['defaults'])
                    sub_component['config'] = config

                    # appends the newly found sub component to the previous/parent component
                    parent_component['children'].append(sub_component)

                    # load the new subcomponent tree
                    sub_component_tree = self.load_tree(sub_component)
                    # if tag == 'digit':
                    #     print('DIGIT !!')
                    #     print(tostring(sub_component_tree, encoding='unicode'))
                    self.merge_nonbody_blocks(tree, sub_component_tree)

                    # nest in place the "body" of the declared component
                    parent.remove(element)
                    comp_wb = sub_component_tree.find('worldbody')
                    parent.extend(comp_wb)

                    # further, nests the declared content, in the declared component,
                    # ==> as last child if no "parent" is specified.
                    # ==> (or) inside the body with name @parent
                    #
                    # e.g. <parent><child_to_nest parent='link_name'></child_to_nest></parent>
                    # the child_to_nest is nested as the last child of the element @name='link_name'
                    component_content_elements = list(element)
                    for nested_element in component_content_elements:
                        if 'parent' in nested_element.attrib:
                            # print('PARENT', nested_element.attrib)
                            parentName = nested_element.attrib['parent']
                            new_parent = parent.find(
                                ".//body[@name='" + (sub_component['id'] + ':' + parentName) + "']")
                            if new_parent is None:
                                raise KeyError('Failed to nest ' +
                                               stringify_header(nested_element) + '>,' +
                                               ' body[name=' +
                                               parentName +
                                               '] not being found in ' + sub_component['tag'] + ' template.')
                            # sub_tree.attrib['__base_component__'] = sub_tree['base_component']
                            nested_element.attrib.pop('parent')
                            new_parent.append(nested_element)
                            walk_tree_recur(new_parent, [nested_element], comp, config)
                        else:
                            new_parent = parent  # sub_component_tree.find('worldbody')
                            new_parent.append(nested_element)
                            walk_tree_recur(new_parent, [nested_element], comp, config)

                else:
                    walk_tree_recur(element, list(element), comp, config)

                for attr in RENAME_KEYS_ATTRS_NAMES:
                    if attr in element.attrib:
                        element.attrib[attr] = parent_component['id'] + ':' + element.attrib[attr]

        walk_tree_recur(tree, [tree.find('worldbody')], comp)

        return tree

    def parse_attributes(self, elem, config):
        def set(k, v):
            elem.attrib[k] = v

        try:
            {set(k, self.replace_all(v, config)) for k, v in elem.attrib.items() if "$" in v}
        except Exception as e:
            raise Exception(e.args[0] + ',\nfor ' + stringify_header(elem), e.args[1])

    def replace_all(self, attr, config):
        # not very pretty code, but it works
        # replaces all {expression} -> eval(expressio, None, config) in the attr.

        wrap = {'txt': attr, 'offset': 0}

        # todo improve this regex to pass python code
        # regex = "(\$\{(\w(\w|\+|\.|\*|\ |\\n|\'|[|]|=|/|%)*)\})"
        regex = "\$\{([^}]+)\}"  # proposed by ChatGPT

        def replace(match, o):
            txt = o['txt']
            offset = o['offset']
            k = match.group(2)
            s = match.span(2)

            v = str(eval(k, None, config))
            # if k not in (config or {}):
            #     raise Exception(
            #         'Failed to find config "' + k + '" while parsing "' + attr + '".',
            #         k
            #     )

            o['txt'] = txt[: offset + s[0] - 2] + v + txt[offset + s[1] + 1:]
            # offset is used to compensate the "original" spans
            # for the differences in the string before and after
            o['offset'] += len(v) - (s[1] - s[0]) - 3

        [replace(x, wrap) for x in re.finditer(regex, attr)]
        return wrap['txt']

    def merge_nonbody_blocks(self, tree, sub_component_tree):
        # auxiliary map that maps child->parent element,
        # for component instantiation elements
        # parent_map = {
        #     child_: parent_
        #     for parent_ in tree.iter()
        #     for child_ in parent_
        #     if child_ == element
        #     # if is_component(child_)
        # }

        extendable_blocks = {
            block_name: tree.find(block_name)
            for block_name in EXTENDABLE_BLOCKS_NAMES
        }

        # merge all blocks actuators, sensors, etc
        # with the current component tree
        for block_name in extendable_blocks:
            block = sub_component_tree.find(block_name)
            if block is not None:
                if extendable_blocks[block_name] is not None:
                    extendable_blocks[block_name].extend(block)
                else:
                    if block_name == 'default' or block_name == 'asset':
                        tree.insert(0, block)
                    else:
                        tree.append(block)

        # remove the "component" tag from the current tree
        # merge the new subcomponent tree/body with the current tree
        # print('>> ', element, parent, parent_map[element], parent == parent_map[element])
        # parent_map[element]
        # parent.remove(element)
        # comp_wb = sub_component_tree.find('worldbody')
        # parent.extend(comp_wb)

        # return parent_map[element]



import mujoco
import cv2
import time
import numpy as np
# from gym.envs.mujoco.mujoco_rendering import RenderContextOffscreen
import os
from .utils.phong_render import PhongRender

gel_width, gel_height = 240,320

def get_simapproach():
    # load the image taken from a real digit sensor

    __location__ = os.path.realpath(
        os.path.join(os.getcwd(), os.path.dirname(__file__)))

    model_path      = os.path.join(__location__, './assets/gel/digit_bg.npy')
    background_img  = np.load(model_path)
    # Phong render params
    ## sensor light sources
    light_sources_digit = [
        {'position': [-0.866, 0.5, 0.25], 'color': (108, 82, 255), 'kd': 0.1, 'ks': 0.4},  # red, bottom (108, 82, 255) kd0.6
        {'position': [0.866, 0.5, 0.25], 'color': (120, 255, 153), 'kd': 0.1, 'ks': 0.4},  # green, left (120, 255, 153)
        {'position': [0, -1, 0.25], 'color': (255, 130, 115), 'kd': 0.3, 'ks': 0.4},  # blue, left (255, 130, 115)
    ]

    ## calculation attributes
    ka                  = 1.0
    px2m_ratio          = 5.4347826087e-05
    elastomer_thickness = 1.0  
    min_depth           = 0 
    texture_sigma       = 0.00001 

    simulation = PhongRender(
        light_sources       = light_sources_digit,
        background_img      = background_img,
        ka                  = ka,
        texture_sigma       = texture_sigma,
        px2m_ratio          = px2m_ratio,
        elastomer_thickness = elastomer_thickness,
        min_depth           = min_depth,
    )

    return simulation

if __name__=="__main__":
    filepath = "digit.xml"

    model = mujoco.MjModel.from_xml_path(filepath)
    data = mujoco.MjData(model)
    viewer = RenderContextOffscreen(gel_width,gel_height,model,data)

    # while True:
    mujoco.mj_step(model,data)
    # obtain depth and mask
    viewer.render(gel_width,gel_height,camera_id=0)
    depth = viewer.read_pixels(gel_width, gel_height, depth=True)[1].copy()
    depth = np.fliplr(depth)

    # render depth to tactile img
    simulation = get_simapproach()
    tact_img = simulation.generate(depth)

    cv2.imshow("tactile_img", tact_img)
    if cv2.waitKey(0) & 0xFF == ord('q'):
        cv2.destroyAllWindows()


from yarok import component, interface
from yarok.platforms.mjc import InterfaceMJC

from yarok.comm.utils.PID import PID
from math import pi


@interface()
class Robotiq2f85MJCInterfaceMJC:

    def __init__(self, interface: InterfaceMJC):
        self.interface = interface
        self.target_q = 0
        self.q = 0
        # P = 0.001
        # I = 0
        # D = 0
        # self.gear = 100
        # self.PIDs = [PID(P, I, D) for a in interface.actuators]
        # self.move(0.9)
        #
        # self.last_query_position = None
        # self.at_target_hit_counter = 0

    def at(self):
        return self.q

    def is_at(self, q):
        return abs(self.q - self.target_q) <= 1

        # def similar_q(a1, a2):
        #     return abs(a1 - a2) < 0.02
        #
        # at = self.at()
        #
        # if self.last_query_position is not None:
        #     if similar_q(a, self.last_query_position):
        #         self.at_target_hit_counter += 1
        #
        #         if self.at_target_hit_counter > 30:
        #             self.last_query_position = None
        #             self.at_target_hit_counter = 0
        #             return True
        #
        #     else:
        #         self.last_query_position = None
        #         self.at_target_hit_counter = 0
        #
        # self.last_query_position = at
        # return False
        pass

    def move(self, q):
        self.target_q = q
        return lambda: self.is_at(q)
        # k = 3.14
        # q = [-1 * a * k * self.gear, a * k * self.gear]
        # [self.PIDs[qa].setTarget(q[qa] * self.gear) for qa in range(2)]
        # pass

    def close(self, percentage=1.0):
        self.move(round(percentage * 255))

    def step(self):
        self.q = round(self.interface.sensordata()[0] * (255 / 0.8))
        # scale =  255 / (0.8 - 0.002)
        # self.q =
        # print(self.q, self.target_q )
        # self.q = data
        # print(self.q)
        # [self.PIDs[a].update(data[a]) for a in range(2)]
        self.interface.set_ctrl(0, self.target_q)
        # print('')

        # m  = ['finger0_joint1_vel', 'finger1_joint1_vel']
        # [self.interface.set_ctrl(m[a], abs(self.PIDs[a].SetPoint - data[a])) for a in range(2)]


@component(
    tag='robotiq-2f85',
    defaults={
        'interface_mjc': Robotiq2f85MJCInterfaceMJC,
        'left_tip': True,
        'right_tip': True
    },
    template_path='2f85.xml',
)
class Robotiq2f85:

    def __init__(self):
        pass

    def is_open(self):
        pass

    def is_closed(self):
        pass

    def is_at(self, a):
        pass

    def move(self, a):
        pass

    def close(self, percentage=1.0):
        pass

    def open(self):
        self.close(0)



from yarok import Platform, PlatformMJC, component, Injector

from yarok.comm.worlds.empty_world import EmptyWorld
from yarok.comm.components.robotiq_2f85.robotiq_2f85 import Robotiq2f85
from yarok.comm.components.ur5e.ur5e import UR5e
from yarok.comm.components.digit.digit import Digit

from yarok.comm.plugins.cv2_inspector import Cv2Inspector

from math import pi


@component(
    extends=EmptyWorld,
    components=[
        UR5e,
        Robotiq2f85,
        Digit
    ],
    # language=xml
    template="""
        <mujoco>
            <worldbody>
               <!--
                <body>
                    <geom type="box" zaxis="1 1 1" size="0.01 0.01 0.01" pos="0.025 0.003 0.157"/>
                    <geom type="box" zaxis="-1 -1 -1" size="0.01 0.01 0.01" pos="-0.03 0.003 0.157"/>
                </body> -->
                <body pos='0 0 0.2'>
                    <freejoint/>
                    <geom type='cylinder' size='0.02 0.11' pos='-0.135 -0.48 0.11' mass='0.0001' zaxis='1 0 0'/>
                </body>
                <body>
                    <geom type="box" size="0.02 0.1 0.09" pos="-0.2 -0.48 0.09"/>
                    <geom type="box" size="0.02 0.1 0.09" pos="-0.05 -0.48 0.09"/>
                    <geom type="box" size="0.01 0.01 0.13" pos="-0.18 -0.44 0.13"/>
                    <geom type="box" size="0.01 0.01 0.13" pos="-0.18 -0.52 0.13"/>
                    <geom type="box" size="0.01 0.01 0.13" pos="-0.07 -0.52 0.13"/>
                    <geom type="box" size="0.01 0.01 0.13" pos="-0.07 -0.44 0.13"/>
                </body>
            <!-- body wrapper alignment for "old" gripper 
                right: xyaxes="0 -1 0 0 0 1" pos="0.0765 0.161 0.092" 
                left: xyaxes="0 -1 0 0 0 -1" pos="-0.0765 0.161 0.092"
            -->
                <ur5e name='arm'> 
                   <robotiq-2f85 name="gripper" parent="ee_link"> 
                        <body xyaxes='0 0 -1 1 0 0' pos="-0.001  0.009 .064" parent='right_tip'>
                            <digit name="digit1" />
                        </body>
                        <body xyaxes='0 0 -1 1 0 0' pos="-0.001  0.009 .064" parent='left_tip'>
                            <digit name="digit2" />
                        </body>
                    </robotiq-2f85> 
                 </ur5e> 
            </worldbody>        
        </mujoco>
    """
)
class DigitUR5World:
    pass


class DigitPickAndPlace:

    def __init__(self, pl: Platform, injector: Injector):
        self.arm: UR5e = injector.get('arm')
        self.gripper = injector.get('gripper')
        self.pl = pl

    def on_update(self):
        q = [0, -pi / 2, -pi / 2 + pi / 4, 0, pi / 2, pi / 2]
        self.arm.move_q(q)
        self.gripper.move(0)
        self.pl.wait(lambda: self.arm.is_at(q) and self.gripper.is_at(0))
        print('----')

        q = [0, -pi / 2, -pi / 2, -pi / 2, pi / 2, pi / 2]
        self.arm.move_q(q)
        self.pl.wait(lambda: self.arm.is_at(q))
        print('----')

        # self.gripper.move(0.78)

        # self.pl.wait_seconds(10)

        q = [0, -pi / 2, - pi / 2 - pi / 10, -pi / 2 + pi / 10, pi / 2, pi / 2]
        self.arm.move_q(q)

        self.pl.wait(lambda: self.arm.is_at(q))
        self.pl.wait_seconds(3)

        self.pl.wait_seconds(5)
        self.gripper.move(140)
        # self.pl.wait(lambda: self.gripper.is_at(200))
        self.pl.wait_seconds(3)

        q = [0, -pi / 2, - pi / 2 - pi / 10 + pi / 5, -pi / 2 + pi / 10 - pi / 5, pi / 2, pi / 2]
        self.arm.move_q(q)
        self.pl.wait(lambda: self.arm.is_at(q))

        q = [0, -pi / 2, -pi / 2 + pi / 4, 0, pi / 2, pi / 2]
        self.arm.move_q(q)
        self.pl.wait(lambda: self.arm.is_at(q))

        # self.gripper.move(0.5)

        # print('open')
        # self.gripper.move(0)
        # self.pl.wait_seconds(5)
        # print('close')
        # self.gripper.move(1)
        # self.pl.wait_seconds(5)


conf = {
    'world': DigitUR5World,
    'behaviour': DigitPickAndPlace,
    'defaults': {
        'environment': 'sim',
        'components': {
            '/gripper': {
                'left_tip': False,
                'right_tip': False,
            }
        },
        'plugins': [
            (Cv2Inspector, {})
        ]
    },
    'environments': {
        'sim': {
            'platform': {
                'class': PlatformMJC,
                'width': 1000,
                'height': 800,
            }
        }
    },
}

if __name__ == '__main__':
    Platform.create(conf).run()



from xml.etree.ElementTree import tostring

import mujoco
from mujoco import MjModel

from yarok.core.components_manager import ComponentsManager
from yarok.core.injector import Injector

from ..mjc.interface import InterfaceMJC
from ..mjc.viewer import ViewerMJC
from yarok.core.config_block import ConfigBlock
from yarok.core.platform import Platform
import numpy as np
import cv2
import inspect

INTERFACES_COMPONENTS_NAMES = [
    'sensor',
    'actuator',
    'cam'
]


def adr2name(model, adr):
    return model.names[adr:].decode().split("\x00")[0]


class PlatformMJC(Platform):
    """
        Is the Platform for the MuJoCo simulator

        Manages the environment platform and necessary integrations.

        1. Starts MjSim with the manager.xml_tree
        2. Instantiates the interfaces for each MuJoCo sensor, actuator or camera.
        3. Instantiates the viewer (i.e. ViewerMJC)
        4. Instantiates the components

        Exposes the step() method that steps() through all the interfaces.
    """

    def __init__(self, manager: ComponentsManager, config: ConfigBlock, behaviour):
        super(PlatformMJC, self).__init__(manager, config, behaviour)

        self.model = MjModel.from_xml_string(tostring(manager.xml_tree, encoding="unicode"))
        self.data = mujoco.MjData(self.model)

        self.manager = manager
        self.viewer = ViewerMJC(self, self.config)

        # self.context = MjRenderContextOffscreen(self.sim)

        interfaces_mjc = {id: InterfaceMJC(id, self)
                          for id, c in self.manager.components.items()
                          if 'interface_mjc' in self.manager.config(id)}

        self.init_mjc_interfaces(interfaces_mjc)

        [interf.on_init() for _, interf in interfaces_mjc.items()]

        # ** {'config': config['interfaces'][c['name_path']] or ConfigBlock({})}
        def init_interface(n):
            interface_cls = self.manager.config(n)['interface_mjc']

            # backwards compatible:
            if not hasattr(self.manager.config(n)['interface_mjc'], '__data__'):
                print('[deprecated] interfaces should be annotated with the @interface decorator')
                return interface_cls(interfaces_mjc[n])

            class_members = {t[0]: t[1] for t in inspect.getmembers(interface_cls)}
            init_members = {t[0]: t[1] for t in inspect.getmembers(class_members['__init__'])}
            init_annotations = init_members['__annotations__'] if '__annotations__' in init_members else {}

            interface_defaults = self.manager.config(n)['interface_mjc'].__data__['defaults']
            config = ConfigBlock({})
            config.defaults(interface_defaults)

            providers = [
                {'match_type': ConfigBlock, 'value': config},
                # {'match_all': true, 'factory': lambda name, type: self.platform.manager.get_by_name(self.component_ref, name)['instance']}
                {'match_type': InterfaceMJC, 'value': interfaces_mjc[n]}
            ]

            injector = Injector(self, providers=providers)

            return interface_cls(**injector.get_all(init_annotations))

        self.interfaces = {n: init_interface(n) for n, interface_mjc in interfaces_mjc.items()}

        self.init_components(self.interfaces)

    def init_mjc_interfaces(self, interfaces_mjc):
        """
            Instantiates interfaces for sensors, actuators, and cameras.
        """

        for item in INTERFACES_COMPONENTS_NAMES:
            # print((self.model.names.decode('utf-8')))
            # print((self.model.nnames))

            for idx, adr in enumerate(getattr(self.model, 'name_' + item + 'adr')):
                abs_name = adr2name(self.model, adr)  # getattr(self.model, item + '_id2name')(idx)
                col_idx = abs_name.index(':')
                component_name = abs_name[:col_idx]
                item_name = abs_name[col_idx + 1:]

                if component_name in interfaces_mjc:
                    interface = interfaces_mjc[component_name]
                    getattr(interface, item + '_name2idx')[item_name] = idx
                    getattr(interface, item + '_name2adr')[item_name] = adr
                    getattr(interface, item + 's').append(item_name)
                    interface.component_name = component_name

    def step(self):
        """
            Steps through all the interfaces
        """
        mujoco.mj_step(self.model, self.data)
        _ = {self.interfaces[i].step() for i in self.interfaces if hasattr(self.interfaces[i], 'step')}
        self.viewer.step()



from yarok import component, interface, ConfigBlock
from yarok.platforms.mjc import InterfaceMJC


@interface(
    defaults={
        'resolution': (640, 480)
    }
)
class CameraInterfaceMJC:

    def __init__(self, mjc: InterfaceMJC, config: ConfigBlock):
        self.interface = mjc
        self.resolution = config['resolution']

    def read(self, depth=False, shape=None):
        shape = (self.resolution[1], self.resolution[0]) if shape is None else shape
        return self.interface.read_camera('cam', shape=shape, depth=depth, rgb=True)


@component(
    defaults={
        'interface_mjc': CameraInterfaceMJC,
        'probe': lambda c: {'camera': c.read()}
    },
    # language=xml
    template="""
        <mujoco>
            <asset>
                <material name="black_plastic" rgba=".3 .3 .3 1"/>
                <material name="gray_ring" rgba="1 1 1 1"/>
            </asset>
            <worldbody>
                <body>
                    <geom type='box' size='0.08 0.04 0.02' pos='0 0 -0.005' material='black_plastic'/>
                    <geom type='cylinder' size='0.02 0.02' material='gray_ring'/>
                    <camera name="cam" zaxis='0 0 -1'/>
                </body>
            </worldbody>
        </mujoco>
    """
)
class Cam:

    def __init__(self):
        pass

    def read(self, depth=False, shape=(480, 640)):
        # Implemented by the interface
        pass


if __name__ == '__main__':
    from yarok import Platform
    from yarok.comm.worlds.empty_world import EmptyWorld
    from yarok.comm.plugins.cv2_inspector import Cv2Inspector


    @component(
        extends=EmptyWorld,
        components=[
            Cam
        ],
        # language=xml
        template="""
            <mujoco>
                <worldbody>
                    <body pos='0 0 0.5' zaxis='0 0 -1'>
                        <cam name='test_cam'/>
                    </body>
                </worldbody>
            </mujoco>
        """
    )
    class CamTestWorld:
        pass


    Platform.create({
        'world': CamTestWorld,
        'defaults': {
            'plugins': [
                (Cv2Inspector, {})
            ]
        }
    }).run()
